import madmike.util.Log;


import java.lang.reflect.InvocationTargetException;


public class Composer
{
    /* --- The rules we follow composing a score --- */
    Ruleset ruleSet;


    private Inspiration numberGenerator;  // = new FileNumberGenerator ("4_2.ack");

    private int numVoices;
    private int maxLength;

    private int currentVoice = 0;  // voice 0 is cello
    private int currentChord = 0;

    /**
     * Here, we store the notes[][] array from the score for performance
     * reasons, so we don't have to use mutator / accessor methods all the
     * time.
     * Voice index is first, chord index is second.
     */
    private int [] [] notes;

    private boolean accept;

    /* --- These are internal state variables for composing. --- */
    private boolean tritoneResolution = false;
    private int tritoneVoice1, tritoneVoice2;

    /** Cantus Firmus climax */
    private int cfClimax;

    private boolean inCadence = false;
    private int leadingToneVoice;


    public Composer() {}


    public Composer(Score s) {
	numVoices = s.getNumVoices();
	maxLength = s.getLength();
	// notes = new int [numVoices] [maxLength];
	notes = s.getNotes();

	numberGenerator = new RandomNumberGenerator ();
    }


    @Deprecated public Composer (Score s, String ruleSetStr, String inspClassName, String [] inspParms)
    {
	numVoices = s.getNumVoices();
	maxLength = s.getLength();
	// notes = new int [numVoices] [maxLength];
	notes = s.getNotes();

	try
	{
	    // Deprecated:
	    /*
	    Class inspClass = Class.forName (inspClassName);
	    numberGenerator = (Inspiration) inspClass.newInstance ();
	    */
	    Class<?> inspClass = Class.forName (inspClassName);
	    numberGenerator = (Inspiration) inspClass.getDeclaredConstructor().newInstance ();
	    numberGenerator.setParms (inspParms);
	}
	catch (ClassNotFoundException ex)
	{
	    Log.write (Log.ERROR, "Class '" + inspClassName + "' not found!");
	}
	catch (NoSuchMethodException ex)
	{
	    Log.write (Log.ERROR, "Class '" + inspClassName + "' does not have the right constructor!");
	}
	catch (InstantiationException ex)
	{
	    Log.write (Log.ERROR, "Class '" + inspClassName + "' cannot be instantiated!");
	}
	catch (IllegalAccessException ex)
	{
	    Log.write (Log.ERROR, "Class '" + inspClassName + "' cannot be accessed!");
	}
	catch (InvocationTargetException ex)
	{
	    Log.write (Log.ERROR, "InvocationTargetException in class '" + inspClassName + "': " + ex.toString ());
	}
    }


    // For future use? (I don't really remember why I wrote this...)
    /*
    public static Score compose(int v, int l, String ruleSetStr, String inspClassName, String[] inspParms) {
	Score sc = new Score(v, l);
	Composer co = new Composer(sc, ruleSetStr, inspClassName, inspParms);
	sc.setRuleSetStr(ruleSetStr.toUpperCase());

	co.compose(sc, ruleSetStr, inspClassName, inspParms);

	return sc;
    }
    */


    /* === BEGIN composition routines === */
    /* --- Compose a score: Try until we succeed --- */
    public void compose(Score sc, String ruleSetStr, String inspClassName, String[] inspParms) {
	numVoices = sc.getNumVoices();
	maxLength = sc.getLength();
	notes = sc.getNotes();
	// Rulesets.setRules(this, ruleSetStr);
	// sc.setRuleSetStr(ruleSetStr.toUpperCase());
	ruleSet = new Ruleset(ruleSetStr.toUpperCase());
	sc.setRuleSetStr(ruleSetStr.toUpperCase());
	
	System.out.println ("Composing " + numVoices + "-part counterpoint over " + maxLength + " chords (using a " + inspClassName + ")...");
	System.out.println("Using Rule Set " + ruleSetStr.toUpperCase() + ".");

	try
	{
	    // Deprecated:
	    /*
	    Class inspClass = Class.forName (inspClassName);
	    numberGenerator = (Inspiration) inspClass.newInstance ();
	    */
	    Class<?> inspClass = Class.forName (inspClassName);
	    numberGenerator = (Inspiration) inspClass.getDeclaredConstructor().newInstance ();
	    numberGenerator.setParms (inspParms);
	}
	catch (ClassNotFoundException ex)
	{
	    Log.write (Log.ERROR, "Class '" + inspClassName + "' not found!");
	}
	catch (NoSuchMethodException ex)
	{
	    Log.write (Log.ERROR, "Class '" + inspClassName + "' does not have the right constructor!");
	}
	catch (InstantiationException ex)
	{
	    Log.write (Log.ERROR, "Class '" + inspClassName + "' cannot be instantiated!");
	}
	catch (IllegalAccessException ex)
	{
	    Log.write (Log.ERROR, "Class '" + inspClassName + "' cannot be accessed!");
	}
	catch (InvocationTargetException ex)
	{
	    Log.write (Log.ERROR, "InvocationTargetException in class '" + inspClassName + "': " + ex.toString ());
	}
	
	int retries = 0;
	long startTime = System.nanoTime ();

	// Generate dynamics "two at a time", p. 156;
	Log.write(Log.INFORMATION, "Generating dynamics...");
	boolean[] hasDynamics = new boolean[numVoices];
	boolean allVoicesHaveDynamics = false;
	int dv = 0;
	for(int i=0; i<numVoices; ++i)
	    hasDynamics[i] = false;
	while(!allVoicesHaveDynamics)
	{
	    int d = numberGenerator.generateValue(4);
	    // Log.write(Log.INFORMATION, "Dyn. value = " + d);
	    int i=0;
	    while(i<2 && !allVoicesHaveDynamics) {
		dv = numberGenerator.generateValue(numVoices);
		if(!hasDynamics[dv]) {
		    sc.setDynamics(dv, d);
		    hasDynamics[dv] = true;
		    ++i;
		}
		allVoicesHaveDynamics = true;
		for(int j=0; j<numVoices; ++j)
		    allVoicesHaveDynamics = allVoicesHaveDynamics && hasDynamics[j];
	    }
	}
	// Debug output
	for(int i=0; i<numVoices; ++i) {
	    Log.write(Log.DEBUG, "Dynamics for voice " + i + " is " + sc.getDynamics(i) + ".");
	}
	
	while (!generate ())
	{
	    // System.out.println (sc.toString ());

	    retries++;
	    Log.write (Log.INFORMATION,            "No success, retrying (" + retries + ")...\n" +
                         	        "           ----------------------------------------------");

	    if (retries == 20000)
		System.out.println ("Please be patient, I'm doing my best...");
	}
	long estimatedTime = System.nanoTime () - startTime;
	estimatedTime /= 1000000;
	
	System.out.println ("Done (" + (retries == 0 ? "at the first attempt" : "after " + retries + " retries") + "), " + 
			    "composing time: " + (int) (((float) (estimatedTime / 1000.0)) * 10) / 10.0 + "s.\n");
    }
    

    /* --- This is where the magic happens --- */
    public boolean generate ()
    {
	Log.write (Log.INFORMATION, "Generating a new score...");
	
	currentVoice = 0;
	currentChord = 0;
	accept = true;
	cfClimax = 0;
	tritoneResolution = false;
	inCadence = false;
	for (int j=0; j<numVoices; j++) for (int i=0; i<maxLength; i++) notes [j] [i] = 0;

	// For the new, not originally H-I backstepping routine:
	int chordsToClear = 1;
	int backstepCounter = 0;

	// If we're not only writing a cadence...
	if (currentChord < maxLength - 2 && ruleSet.getRule(Ruleset.RULE_INITIALTONIC))
	    setInitialNotes ();

	// Move along chords...
	for (; currentChord < maxLength && accept; currentChord++)
	{
	    // --------------------------------------------------
	    //  As we can get out of cadence by the new backstepping routine:
	    inCadence = currentChord >= maxLength - 2 && ruleSet.getRule(Ruleset.RULE_CADENCE);
	    // --------------------------------------------------
	    // if (currentChord == maxLength - 2 && ruleCadence) inCadence = true;

	    int retries = 0;

	    do
	    {
		Log.write (Log.DEBUG, "Setting chord #" + currentChord + ", retry #" + retries + "...");

		accept = true;

		// For every voice...
		for (currentVoice = 0; currentVoice < numVoices && accept; currentVoice++)
		{
		    // Tritone resolution.
		    if (tritoneResolution && ruleSet.getRule(Ruleset.RULE_TRITONERESOLUTION))
		    {
			if (currentVoice == tritoneVoice1 || currentVoice == tritoneVoice2)
			{
			    Log.write (Log.DEBUG, "Tritone resolution in voice " + currentVoice + "...");

			    switch (notes [currentVoice] [currentChord - 1])
			    {
				case 3:
				    notes [currentVoice] [currentChord] = 2;
				    break;
				case 6:
				    notes [currentVoice] [currentChord] = 7;
				    break;
				case 10:
				    notes [currentVoice] [currentChord] = 9;
				    break;
				case 13:
				    notes [currentVoice] [currentChord] = 14;
				    break;
			    }
			}
		    }
		    else  // no tritone resolution
		    {
			// No skip-stepwise motion.
			if (skipStepwise () >= 0 || !ruleSet.getRule(Ruleset.RULE_SKIPSTEPWISE))
			{
			    // Last chord of cadence is tonic triad in root position.
			    if (inCadence && currentChord == maxLength - 1)
			    {
				if (currentVoice == 0)
				    notes [currentVoice] [currentChord] = 7 * numberGenerator.generateValue (2);  // 7 * (int) (2 * Math.random ());
				else
				    notes [currentVoice] [currentChord] = Score.tonicTriad [numberGenerator.generateValue (Score.tonicTriad.length)];  // tonicTriad [(int) (tonicTriad.length * Math.random ())];
			    }
			    else
			    {
				// Generate random note.
				generateNote ();
			    }
			}
			else  // skip-stepwise motion.
			{
			    Log.write (Log.DEBUG, "Stepwise motion required in voice " + currentVoice + ".");

			    int step = numberGenerator.generateValue (3) - 1; // (int) (Math.random () * 3) - 1;
			    notes [currentVoice] [currentChord] = notes [currentVoice] [currentChord - 1] + step;
			    if (notes [currentVoice] [currentChord] < 0 || notes [currentVoice] [currentChord] > 14)
			    {
				Log.write (Log.DEBUG, "Unacceptable note in stepwise motion!");
				accept = false;
			    }
			}
		    }

		    // Three-note repeat.
		    if (accept && threeNoteRepeat () && ruleSet.getRule(Ruleset.RULE_FORBIDTHREENOTEREPEAT))
		    {
			Log.write (Log.DEBUG, "Three-note repeat!");
			accept = false;
		    }

		    // Melodic subroutine.
		    if (accept && ruleSet.getRule(Ruleset.RULE_MELODIC)) checkMelody ();
		    if (accept && ruleSet.getRule(Ruleset.RULE_OCTAVERANGE)) checkOctaveRange ();

		}  // END voice count loop.

		// Harmonic subroutine.  TODO: Nested loops.
		if (accept && ruleSet.getRule(Ruleset.RULE_HARMONIC))
		{
		    for (int j=0; j<numVoices && accept; j++)
			for (int i=0; i<numVoices && accept; i++)
			{
			    if (i != j)
			    {
				Log.write (Log.DEBUG, "Checking harmony between voices " + i + " and " + j + "...");
				checkHarmony (i, j);
				Log.write (Log.DEBUG, "Checking harmony between voices " + j + " and " + i + "...");
				checkHarmony (j, i);
			    }
			}
		}

		// At least one voice stepwise.
		// Extension of H-I: It makes no sense if only one voice!
		if (accept && ruleSet.getRule(Ruleset.RULE_ONE_STEPWISE) && numVoices > 1 && currentChord > 0)
		{
		    boolean stepwise = false;
		    for (int j=0; j<numVoices && !stepwise; j++)
		    {
			stepwise = (Math.abs (notes [j] [currentChord] - notes [j] [currentChord - 1]) <= 1);
		    }
		    if (!stepwise)
		    {
			Log.write (Log.DEBUG, "No stepwise motion in chord " + currentChord + "!");
			accept = false;
		    }
		}

		// No 6-4-chords.
		if (accept && ruleSet.getRule(Ruleset.RULE_FORBID64CHORDS))
		{
		    // Find lowest note, cf. p. 107.
		    int lowestNote = 0;
		    int lowestVoice = 0;
		    for (int j=0; j<numVoices; j++)
			if (notes [j] [currentChord] <= lowestNote)
			{
			    lowestNote = notes [j] [currentChord];
			    lowestVoice = j;
			}

		    for (int j=0; j<numVoices && accept; j++)
		    {
			if (j != lowestVoice)
			{
			    if (notes [j] [currentChord] - lowestVoice == 3 || notes [j] [currentChord] - lowestVoice == 10)
			    {
				StringBuffer b = new StringBuffer ();
				for (int k=0; k<numVoices; k++)
				{
				    b.append ("V");
				    b.append (k);
				    b.append (": ");
				    b.append (Score.noteLabels [notes [k] [currentChord]]);
				    b.append ("  ");
				}
				Log.write (Log.DEBUG, "6-4-chord found at position " + currentChord + ": " + b.toString () + "!");
				accept = false;
			    }
			}
		    }
		}

		if (accept && ruleSet.getRule(Ruleset.RULE_TRITONERESOLUTION)) setTritoneResolution ();

		if (accept && inCadence)
		{
		    // Exactly one leading tone in cadence.
		    if (currentChord == maxLength - 2)
		    {
			int countLeadingTones = 0;
			// Log.write (Log.INFORMATION, "Counting leading tones...");
			for (int j=0; j<numVoices; j++)
			{
			    if (notes [j] [currentChord] == 6 || notes [j] [currentChord] == 13)
			    {
				countLeadingTones++;
				// Log.write (Log.INFORMATION, "Leading tone in voice " + j + ".");
				leadingToneVoice = j;
			    }
			}
			if (countLeadingTones != 1)
			{
			    Log.write (Log.DEBUG, "Not exactly one leading tone (but " + countLeadingTones + ") in next-to-last chord!");
			    accept = false;
			}
		    }
		    // Last chord in cadence is tonic triad in root position with resolved leading tone.
		    else if (currentChord == maxLength - 1)
		    {
			if (notes [leadingToneVoice] [currentChord] != notes [leadingToneVoice] [currentChord - 1] + 1)
			{
			    Log.write (Log.DEBUG, "Leading tone in voice " + leadingToneVoice + " not resolved!");
			    accept = false;
			}

			if (notes [0] [currentChord] != 0 && notes [0] [currentChord] != 7)
			{
			    Log.write (Log.DEBUG, "Lowest voice in last chord not C-0 or C-1!");
			    accept = false;
			}

			boolean lastChordOk = true;
			for (int j=1; j<numVoices && lastChordOk; j++)
			{
			    boolean isTonic = false;
			    for (int i=0; i<Score.tonicTriad.length && !isTonic; i++)
				isTonic |= notes [j] [currentChord] == Score.tonicTriad [i];
			    lastChordOk &= isTonic;
			}
			if (!lastChordOk)
			{
			    Log.write (Log.DEBUG, "Last chord is not tonic triad!");
			    accept = false;
			}
		    }
		}

		// Test for climax of highest voice.
		if (accept && ruleSet.getRule(Ruleset.RULE_FORBIDCLIMAXREPEAT))
		{
		    if (!tritoneResolution && !inCadence && notes [numVoices - 1] [currentChord] == cfClimax)
		    {
			Log.write (Log.DEBUG, "Illegal repeat of climax!");
			accept = false;
		    }

		    if (notes [numVoices - 1] [currentChord] > cfClimax)
		    {
			cfClimax = notes [numVoices - 1] [currentChord];
			Log.write (Log.DEBUG, "cfClimax = " + cfClimax + " in chord " + currentChord + ".");
		    }
		}

		// Set retry if not accepted.
		if (!accept)
		{
		    retries++;
		    Log.write (Log.DEBUG, "Chord unacceptable, retry #" + retries + "...");
		}

	    } while (!accept && retries < 50);  // cf. H-I, p. 89

	    if (!accept)
	    {
		// --------------------------------------------------
		//  NEW backstepping routine, not original H-I
		accept = true;
		backstepCounter++;
		if (backstepCounter >= 50)
		{
		    chordsToClear++;
		    backstepCounter = 0;
		}
		if (chordsToClear > 3) chordsToClear = 3;  // This is a backstep by 2.
		Log.write (Log.DEBUG, "Clearing " + chordsToClear + " chords at position " + currentChord + "...");
		currentChord -= chordsToClear;
		if (currentChord < 0)
		{
		    Log.write (Log.INFORMATION, "Score rejected!");
		    return false;
		}
		setTritoneResolution ();
		if (tritoneResolution) Log.write (Log.DEBUG, "Tritone after backstepping to chord " + currentChord + "...");
		// --------------------------------------------------
		// This is original H-I, or at least as clear as it gets from their book:
		/*
		Log.write (Log.INFORMATION, "Score rejected at chord position " + currentChord + "!");
		return false;
		*/
	    }
	}  // END chord count loop.

	numberGenerator.dispose ();

	Log.write (Log.INFORMATION, "Score accepted.");
	return true;
    }  // END generate

    private void setInitialNotes ()
    {
	currentVoice = 0;
	currentChord = 0;

	notes [currentVoice++] [currentChord] = 7;

	for (; currentVoice < numVoices - 1; currentVoice++)
	    notes [currentVoice] [currentChord] = Score.tonicTriad [numberGenerator.generateValue (7)];  // tonicTriad [(int) (Math.random () * 7)];

	if (currentVoice < numVoices)
	    notes [currentVoice++] [currentChord] = 7 * numberGenerator.generateValue (3);  // 7 * (int) (3 * Math.random ());

	cfClimax = notes [numVoices - 1] [currentChord];

	currentChord++;
    }  // END setInitialNotes

    private int skipStepwise ()
    {
	if (currentChord < 2) return 1;

	return (-Math.abs (notes [currentVoice] [currentChord - 1] - notes [currentVoice] [currentChord - 2]) + 1);
    }  // END skipStepwise

    private boolean threeNoteRepeat ()
    {
	if (currentChord < 2) return false;

	return (notes [currentVoice] [currentChord - 2] == notes [currentVoice] [currentChord - 1]
	     && notes [currentVoice] [currentChord - 1] == notes [currentVoice] [currentChord]);
    }  // END threeNoteRepeat

    private void checkMelody ()
    {
	if (currentChord == 0) return;

	int n1 = notes [currentVoice] [currentChord - 1];
	int n2 = notes [currentVoice] [currentChord];
	int delta = Math.abs (n1 - n2);
	// delta -= 7;

	boolean checkTritone = false;

	if (delta >= 7)
	{
	}
	else
	{
	    if (delta == 6)
	    {
		accept = false;
		return;
	    }
	    if (delta == 4)
	    {
		checkTritone = true;
	    }
	    else
	    {
		if (delta == 3) checkTritone = true;
	    }
	}

	if (checkTritone)  // 3, 6, 10, 13
	{
	    if ((n1 == 3 && n2 == 6)
	     || (n1 == 6 && n2 == 3)
	     || (n1 == 6 && n2 == 10)
	     || (n1 == 10 && n2 == 6)
	     || (n1 == 10 && n2 == 13)
	     || (n1 == 13 && n2 == 10)
	     || (n1 == 3 && n2 == 13)
	     || (n1 == 13 && n2 == 3))
	    {
		Log.write (Log.DEBUG, "Tritone interval in melody!");
		accept = false;
		return;
	    }
	}
    }  // END checkMelody

    private void checkOctaveRange ()
    {
	// Check octave range.
	for (int i=0; i<currentChord && accept; i++)
	{
	    // "The quantity 8, which is equivalent to the span of a ninth, was then subtracted
	    // from the absolute value of this difference, and if the result remained positive..."
	    // (p. 101) -- This is obviously to be interpreted as "zero or greater"!
	    if (Math.abs (notes [currentVoice] [currentChord] - notes [currentVoice] [i]) - 8 >= 0)
	    {
		Log.write (Log.DEBUG, "Octave-range rule violated!");
		accept = false;
		return;
	    }
	}
    }  // END checkOctaveRange

    private void checkHarmony (int v1, int v2)
    {
	boolean checkParallel = false;
	boolean checkContraryStep = false;

	int delta = Math.abs (notes [v1] [currentChord] - notes [v2] [currentChord]);

	if (delta > 7) delta -= 7;

	switch (delta)
	{
	    case 7:  // Octave
		if (ruleSet.getRule(Ruleset.RULE_FORBIDPARALLELPERFECTS)) checkParallel = true;
		break;
	    case 6:  // Seventh
		accept = false;
		return;
	    case 5:  // Sixth
		checkContraryStep = true;
		break;
	    case 4:  // Fifth
		if (ruleSet.getRule(Ruleset.RULE_FORBIDPARALLELPERFECTS)) checkParallel = true;
		break;
	    case 3:  // Fourth
		if (ruleSet.getRule(Ruleset.RULE_FORBIDPARALLELFOURTHS)) checkParallel = true;
		break;
	    case 2:  // Third
		checkContraryStep = true;
		break;
	    case 1:  // Second
		accept = false;
		return;
	    case 0:  // Unison
		if (ruleSet.getRule(Ruleset.RULE_FORBIDPARALLELPERFECTS)) checkParallel = true;
		break;
	    default:
		Log.write (Log.WARNING, "Unknown harmonic interval at chord " + currentChord + "!   " +
			   "(V" + v1 + ": " + notes [v1] [currentChord] + ", V" + v2 + ": " + notes [v2] [currentChord] + ")\n" +
			   toString () + "!");
		accept = false;
		System.exit (255);
		return;
	}

	if (currentChord > 0)
	{
	    if (checkParallel)
	    {
		int oldDelta = Math.abs (notes [v1] [currentChord - 1] - notes [v2] [currentChord - 1]);
		if (oldDelta > 7) oldDelta -= 7;

		if (delta == oldDelta)
		{
		    Log.write (Log.DEBUG, "Parallel motion (either implicit or explicit)!");
		    accept = false;
		    return;
		}

		checkContraryStep = true;
	    }
	    if (checkContraryStep)
	    {
		int delta1 = notes [v1] [currentChord - 1] - notes [v1] [currentChord];
		int delta2 = notes [v2] [currentChord - 1] - notes [v2] [currentChord];

		if (Math.signum (delta1) == Math.signum (delta2))
		{
		    boolean contraryExists = false;
		    for (int j=0; j<numVoices && !contraryExists; j++)
		    {
			if (j != v1 && j != v2)
			{
			    int delta3 = notes [j] [currentChord - 1] - notes [j] [currentChord];
			    if (Math.signum (delta3) != Math.signum (delta1) && Math.abs (delta3) <= 1)
				contraryExists = true;
			}
		    }
		    if (!contraryExists)
		    {
			Log.write (Log.DEBUG, "No contrary stepwise motion in parallel entering of consonant interval!");
			accept = false;
			return;
		    }
		}
	    }
	}
    }  // END checkHarmony

    private boolean tritoneBetween (int v1, int v2)
    {
	return ((notes [v1] [currentChord] ==  3 && notes [v2] [currentChord] ==  6)
	     || (notes [v1] [currentChord] ==  6 && notes [v2] [currentChord] ==  3)
	     || (notes [v1] [currentChord] ==  3 && notes [v2] [currentChord] == 13)
	     || (notes [v1] [currentChord] == 13 && notes [v2] [currentChord] ==  3)
	     || (notes [v1] [currentChord] ==  6 && notes [v2] [currentChord] == 10)
	     || (notes [v1] [currentChord] == 10 && notes [v2] [currentChord] ==  6)
    	     || (notes [v1] [currentChord] == 10 && notes [v2] [currentChord] == 13)
	     || (notes [v1] [currentChord] == 13 && notes [v2] [currentChord] == 10));
    }  // END tritoneBetween

    private void setTritoneResolution ()
    {
	int numTritones = 0;
	for (int j=0; j<numVoices; j++)
	    for (int i=0; i<j; i++)
	    {
		Log.write (Log.DEBUG, "Checking for tritone between voices " + j + " and " + i + "...");
		if (i != j && tritoneBetween (i, j))
		{
		    numTritones++;
		    tritoneVoice1 = i;
		    tritoneVoice2 = j;
		    Log.write (Log.DEBUG, "Tritone detected between voices " + i + " and " + j + " in chord " + currentChord + ".");
		}
	    }

	switch (numTritones)
	{
	    case 0:
		tritoneResolution = false;
		break;
	    case 1:
		tritoneResolution = true;
		break;
 	    default:
		Log.write (Log.DEBUG, "More than one tritone!");
		accept = false;
	}
    }  // END setTritoneResolution

    private void generateRandomNote ()
    {
	notes [currentVoice] [currentChord] = (int) (Math.random () * 15);
    }  // END generateRandomNote

    private void generateNote ()
    {
	notes [currentVoice] [currentChord] = numberGenerator.generateValue (15);
    }  // END generateNote

    
    /* --- Main Program --- */
    public static void main (String [] args) throws Exception
    {
	int voices = 4;
	int length = 12;
	String ruleSet = "H";
	String inspClassName = "RandomNumberGenerator";
	String [] inspParms = null;
	int loglevel = Log.WARNING;
	boolean silent = true;
	String outfile = "Last.x2s";

	int i=0;

	System.out.println ("Composer: A counterpoint algorithm based on");
	System.out.println ("          ILLIAC Suite Experiment No. 2 (Four-part first-species counterpoint)");
	System.out.println ("          (almost complete, not thoroughly checked, but bearable)");
	System.out.println ("");

	// Check command line arguments
	try
	{
	    if (Integer.parseInt (args [i]) != 0)
		voices = Integer.parseInt (args [i++]);

	    if (Integer.parseInt (args [i]) != 0)
		length = Integer.parseInt (args [i++]);
	}
	catch (Exception ex) {}

	for (; i<args.length; i++)
	{
	    if (args [i].toUpperCase ().equals ("/LOGLEVEL") || args [i].toUpperCase ().equals ("-LOGLEVEL"))
		loglevel = Integer.parseInt (args [++i]);

	    /*
	    if (args [i].toUpperCase ().equals ("/SILENT") || args [i].toUpperCase ().equals ("-SILENT"))
		silent = true;
	    */
	    if (args [i].toUpperCase ().equals ("/PLAY") || args [i].toUpperCase ().equals ("-PLAY"))
		silent = false;

	    if (args [i].toUpperCase ().equals ("/RULESET") || args [i].toUpperCase ().equals ("-RULESET"))
		ruleSet = args [++i];

	    if(args[i].toUpperCase().equals("/O") || args[i].toUpperCase().equals("-O")) {
		outfile = args[++i];
	    }
	    
	    if (args [i].toUpperCase ().equals ("/INSPIRATION") || args [i].toUpperCase ().equals ("-INSPIRATION"))
	    {
		inspClassName = args [++i];
		if (i < args.length - 1)
		{
		    i++;
		    inspParms = new String [args.length - i];
		    for (int j=i; j<args.length; j++)
			inspParms [j-i] = args [j];
		}
	    }

	    if (args [i].toUpperCase ().equals ("/HELP") || args [i].toUpperCase ().equals ("/?")
		|| args [i].toUpperCase ().equals ("-HELP") || args [i].toUpperCase ().equals ("-?"))
	    {
		System.out.println ("Command line format is:");
		System.out.println ("   Composer [[<voices> <length>] [/PLAY] [/O <outfile>] [/LOGLEVEL <l>] [/INSPIRATION <note generator> <parms>] | /HELP | /?]");
		System.out.println ("");
		System.out.println ("   voices     number of voices");
		System.out.println ("   length     length of score");
		System.out.println ("");
		System.out.println ("   l        can give you a glimpse of the internals, range 0--4");
		System.out.println ("");
		System.out.println ("   /PLAY    play the score with the help of the Musician class.");
		System.out.println ("");
		System.out.println ("   outfile  write suite to the file specified");
		System.out.println ("            (default is " + outfile + ")");
		System.out.println ("");
		System.out.println ("   /INSPIRATION sets an alternative note generator.");
		System.out.println ("      parms   whatever parameters it may expect...");
		System.out.println ("");
		System.out.println ("   /HELP or /?   displays this help message");
		System.out.println ("");

		System.exit (0);
	    }
	}

	Log.setLevel (loglevel);

	// Compose a score
	Score sc = new Score (voices, length);
	// Rulesets.setRules (co, ruleSet);
	sc.setRuleSetStr(ruleSet.toUpperCase());
	
	Composer co = new Composer();
	co.compose(sc, ruleSet, inspClassName, inspParms);	

	// Output and store the score
	System.out.println (sc.toString ());
	Log.write(Log.INFORMATION, "Writing score to file " + outfile + " ...");
	Librarian.store (sc, new java.io.File (outfile));

	// Play the score iff not silent
	if (!silent)
	{
	    Musician syn = new Musician ();
	    syn.setMidiDevice(syn.getDeviceFromConf());
	    syn.open ();
	    syn.play (sc);
	    syn.close ();
	}
    }
}
