import madmike.util.Log;


import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.Receiver;
import javax.sound.midi.ShortMessage;
// import javax.sound.midi.Synthesizer;
import javax.sound.midi.Transmitter;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.InvalidMidiDataException;

import javax.sound.midi.Sequencer;
import javax.sound.midi.Sequence;
import javax.sound.midi.MetaMessage;
import javax.sound.midi.MetaEventListener;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.Vector;
import java.util.Arrays;


public class Musician implements MetaEventListener
{
    private static final int NOTE_DURATION = 1000;
    
    // private Synthesizer syn;
    private MidiDevice device;
    private Receiver receiver;
    private Sequencer sequencer;
    private Transmitter transmitter;

    private volatile boolean inLivePlayback = false;  // do not send MIDI messages iff true

    
    /** Shutdown hook: end all notes on early exit */
    private Thread shutdownHook = new Thread() {
	    public void run() {
		Log.write(Log.INFORMATION, "Shutdown hook responding...");

		// Stop sending MIDI messages
		if(inLivePlayback) {
		    inLivePlayback = false;

		    // Wait for last notes to trickle through
		    try {
			Thread.currentThread().sleep(10);
		    }
		    catch(InterruptedException ex) {}
		}

		// Shut down sequencer, if present (MIDI file playback)
		if(sequencer != null) {
		    Log.write(Log.DEBUG, "Stopping and closing sequencer...");
		    sequencer.stop();
		    sequencer.close();

		    transmitter = null;
		    receiver = null;
		}

		// Send all notes off and close receiver
		if(receiver != null) {
		    Log.write(Log.DEBUG, "Sending All Notes Off...");
		    for(int j=0; j<0x10; ++j) {
			// This is safe, cf.
			// https://forum.juce.com/t/midi-all-notes-off-message/10353/3
			// access date: 2023-05-19
			for(int note=0; note<128; ++note) {
			    ShortMessage msg = createShortMessage(ShortMessage.NOTE_OFF | (j & 0x0f), note, 0);
			    receiver.send(msg, 0);
			}
			// This is quick:
			/*
			ShortMessage msg = createShortMessage(0xb0 | (j & 0x0f), 123, 0);
			receiver.send(msg, 0);
			*/
		    }

		    Log.write(Log.DEBUG, "Closing receiver...");
		    receiver.close();
		    receiver = null;
		}

		// Shutdown transmitter, if present (MIDI file playback)
		if(transmitter != null) {
		    Log.write(Log.DEBUG, "Closing transmitter...");
		    transmitter.close();
		}

		// Shut down device
		if(device != null) {
		    device.close();
		    Log.write(Log.DEBUG, "MIDI device closed in shutdown hook.");
		}
	    }
	};

    
    public void meta(MetaMessage meta) {
	if(meta.getType() == 0x01) {
	    System.out.println("Meta text message: " + new String(meta.getData()));
	}
    }


    public static String[] getReceiverDeviceNames() throws MidiUnavailableException {
	MidiDevice.Info[] infos = MidiSystem.getMidiDeviceInfo();
	Vector<String> devNamesVec = new Vector<String>();

	for(int i=0; i<infos.length; ++i) {
	    MidiDevice dev = MidiSystem.getMidiDevice(infos[i]);

	    int n = dev.getMaxReceivers();
	    if(n != 0)
		devNamesVec.add(infos[i].getName().trim());
	}

	// This from https://www.geeksforgeeks.org/how-to-convert-vector-to-array-in-java/
	// access date: 2023-05-19
	Object[] objs = devNamesVec.toArray();
	return Arrays.copyOf(objs, objs.length, String[].class);
    }


    /**
     * Reads the number of the MidiDevice to be used from the .conf file.
     * The number in the .conf file is the index in
     * MidiSystem.getMidiDeviceInfo()[].
     */
    public static MidiDevice getDeviceFromConf() throws FileNotFoundException, IOException, MidiUnavailableException {
	MidiDevice dev;

	String[] deviceNames = getReceiverDeviceNames();
	String devStr = Engineer.getConfString("MIDIDEVICE");
	
	// Read MIDI device from setup file
	int deviceIndex = 0;

	for(int i=0; i<deviceNames.length; ++i)
	    if(devStr.equals(deviceNames[i])) {
		deviceIndex = i;
		break;
	    }

	dev = MidiSystem.getMidiDevice (MidiSystem.getMidiDeviceInfo () [deviceIndex]);
	Log.write (Log.DEBUG, "Device : " + dev.getDeviceInfo ().toString ());

	return dev;
    }


    public void setMidiDevice(MidiDevice dev) {device = dev;}
    

    /* === BEGIN .x2s Score Playback Routines === */
    public void open () throws FileNotFoundException, IOException, MidiUnavailableException, InterruptedException
    {
	Log.write (Log.DEBUG, "Opening MIDI devices...");

	/*
	syn = MidiSystem.getSynthesizer ();
	syn.open ();
	*/

	// device = getDeviceFromConf();

	// Open device && load instruments
	device.open ();
	receiver = device.getReceiver ();

	/*
	for (int i=0; i<syn.getAvailableInstruments ().length; i++)
	    syn.loadInstrument (syn.getAvailableInstruments () [i]);
	*/

	/*
	receiver.send (createShortMessage (ShortMessage.PROGRAM_CHANGE | 0x00, 42, 0), 0);  // Cello
	receiver.send (createShortMessage (ShortMessage.PROGRAM_CHANGE | 0x01, 41, 0), 0);  // Viola
	receiver.send (createShortMessage (ShortMessage.PROGRAM_CHANGE | 0x02, 40, 0), 0);  // Violin
	receiver.send (createShortMessage (ShortMessage.PROGRAM_CHANGE | 0x03, 40, 0), 0);  // Violin
	*/
	
	Thread.currentThread ().sleep (100);
    }

    public void play (Score sc) throws InterruptedException
    {
	int len = sc.getLength();
	int voices = sc.getNumVoices();
	int note, midiNote;

	// For ligatures
	int[] lastNotes = new int[voices];
	int[] notesPlaying = new int[voices];

	for(int i=0; i<voices; ++i)
	    lastNotes[i] = -1;

	/*
	for (int i=0; i<voices; i++)
	    receiver.send (createShortMessage (ShortMessage.PROGRAM_CHANGE | i, 0, 0), 0);  // Piano
	*/

	try {
	    Runtime.getRuntime().addShutdownHook(shutdownHook);
	}
	catch(Exception ex) {}
	
	inLivePlayback = true;
	for (int i=0; i<len && inLivePlayback; i++)
	{
	    // Notes on
	    for (int j=0; j<voices; j++)
	    {
		note = Transcriber.toMidiNote (sc.getNote (j, i));

		// Transpose, depending on voice number.
		midiNote = Transcriber.transposeNoteForVoice(note, j, voices);

		int velocity = Transcriber.dynamicsToMidiVelocity(sc.getDynamics(j));
		// Do a note on iff last note was different.
		if(midiNote != lastNotes[j]) {
		    receiver.send (createShortMessage (ShortMessage.NOTE_ON | (j & 0x0f), midiNote, velocity), 0);
		    notesPlaying[j] = midiNote;
		    
		    // Log.write(Log.DEBUG, "Note ON : " + midiNote + " for voice " + j + ".");
		}
	    }

	    // Hold note durations
	    if(inLivePlayback && !Thread.currentThread().interrupted()) {
		try {
		    Thread.currentThread ().sleep (NOTE_DURATION);
		    if (i == len-1) Thread.currentThread ().sleep ((int)(NOTE_DURATION / 4.0f));
		}
		catch(InterruptedException ex) {}
	    }

	    // Notes off
	    for (int j=0; j<voices; j++)
	    {
		note = Transcriber.toMidiNote (sc.getNote (j, i));

		// Transpose, depending on voice number.
		midiNote = Transcriber.transposeNoteForVoice(note, j, voices);

		// Do a note off iff next note is different, or end of score,
		// or interrupted (i.e. no longer inLivePlayback.
		if(i<len-1 && notesPlaying[j] != Transcriber.transposeNoteForVoice(Transcriber.toMidiNote(sc.getNote(j, i+1)), j, voices) || i == len-1 || !inLivePlayback) {
		    receiver.send (createShortMessage (ShortMessage.NOTE_OFF | (j & 0x0f), midiNote, 0), 0);
		    
		    // Log.write(Log.DEBUG, "Note OFF: " + midiNote + " for voice " + j + ".");
		}
	    }

	    // Remember last note
	    for(int j=0; j<voices; ++j)
		lastNotes[j] = notesPlaying[j];

	    // Pause after playing a score
	    if(inLivePlayback && !Thread.currentThread().interrupted())
		try {
		    if (i == len-1) Thread.currentThread ().sleep ((int)(NOTE_DURATION * 3.0f / 4.0f) + NOTE_DURATION);
		}
		catch(InterruptedException ex) {}
	}

	// We get an IllegalStateException if shutdown is in progress.
	try {
	    Runtime.getRuntime().removeShutdownHook(shutdownHook);
	}
	catch(IllegalStateException ex) {}
    }

    /*
    // private static int [] midiNotes = {60, 62, 64, 65, 67, 69, 71, 72, 74, 76, 77, 79, 81, 83, 84};
    private static int [] midiNotes = {48, 50, 52, 53, 55, 57, 59, 60, 62, 64, 65, 67, 69, 71, 72};
    public static  int toMidiNote (int hiNote) {return midiNotes [hiNote];}
    */

    public void stopPlayback() {
	inLivePlayback = false;
    }  // END stopPlayback

    public void close ()
    {
	if(inLivePlayback) {
	    Runtime.getRuntime().removeShutdownHook(shutdownHook);
	
	    // syn.close ();
	    receiver.close ();
	    device.close ();

	    receiver = null;
	    device = null;
	    
	    Log.write (Log.DEBUG, "MIDI devices closed in method close().");
	}
	inLivePlayback = false;
    }

    private ShortMessage createShortMessage (int status, int data1, int data2)
    {
	ShortMessage msg = new ShortMessage ();
	try {msg.setMessage (status, data1, data2);}
	catch (Exception ex) {Log.write (Log.WARNING, ex.toString ());}
	return msg;
    }
    /* === END .x2s Score Playback Routines === */


    /**
     * Plays the MIDI file, no questions asked.
     * @param midiFile the MIDI file to be played
     */
    private void playMidiFile(File midiFile) throws FileNotFoundException, IOException, MidiUnavailableException, InvalidMidiDataException, InterruptedException {
	sequencer = MidiSystem.getSequencer();
	Sequence s = MidiSystem.getSequence(midiFile);

	// Get device
	// device = getDeviceFromConf();
	device.open();
	
	// Get receiver from device
	receiver = device.getReceiver();
	Thread.currentThread().sleep(100);

	// Connect the receiver as the Sequencer's transmitter
	transmitter = sequencer.getTransmitter();
	transmitter.setReceiver(receiver);

	// Sequencer in action
	sequencer.setSequence(s);

	sequencer.addMetaEventListener(this);
	
	sequencer.open();
	
	Runtime.getRuntime().addShutdownHook(shutdownHook);
	
	sequencer.start();
	while(sequencer.isRunning())
	    Thread.yield();
	
	Thread.currentThread().sleep(1000);

	sequencer.close();
	
	sequencer.removeMetaEventListener(this);

	Runtime.getRuntime().removeShutdownHook(shutdownHook);
	// Sequencer has finished

	receiver.close();
	transmitter.close();
	device.close();
    }


    /**
     * Plays the Score passed.
     * @param sc the score to be played
     */
    public void playScore(Score sc) throws FileNotFoundException, IOException, MidiUnavailableException, InterruptedException {
	Musician mus = new Musician ();
	mus.open ();
	mus.play (sc);
	mus.close ();
    }


    /**
     * Plays the music in the file passed.
     * File type is determined by the ending of the file name.
     * Currently supported: .x2s, .mid
     * @param filename the name of the file to be played back
     */
    public void playFile(String filename) throws FileNotFoundException, IOException, MidiUnavailableException, InvalidMidiDataException, InterruptedException {
	// Play Experiment Two Score
	if(filename.toUpperCase().endsWith(".X2S")) {
	    Score sc = Librarian.fetch (new File (filename));

	    System.out.println("Playing score " + filename + " ...");
	    System.out.println();
	    String ruleSetStr = sc.getRuleSetStr();
	    if(ruleSetStr != null)
		System.out.println("Rule Set " + ruleSetStr);
	    System.out.println (sc.toString ());

	    setMidiDevice(getDeviceFromConf());
	    open();
	    play(sc);
	    close();
	}
	// Play MIDI file
	else if(filename.toUpperCase().endsWith(".MID")) {
	    System.out.println("Playing file " + filename + " ...");

	    // Musician mus = new Musician();
	    playMidiFile(new File(filename));
	}
    }


    /**
     * Main program
     */
    public static void main (String [] args) throws Exception
    {
	System.out.println("Java Musician: Play back .x2s scores or MIDI files");
	System.out.println("");
	
	Log.setLevel(Log.DEBUG);

	Musician mus = new Musician();
	mus.setMidiDevice(getDeviceFromConf());
	mus.playFile(args[0]);
    }
}
