import madmike.util.Log;


import javax.sound.midi.MidiSystem;
import javax.sound.midi.Sequence;
import javax.sound.midi.Track;
import javax.sound.midi.MidiEvent;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.MetaMessage;
import javax.sound.midi.InvalidMidiDataException;

import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;

import java.io.File;
import java.io.IOException;

import java.util.Vector;


/**
 * Class for translating H-I notes to MIDI notes, transposing notes specific
 * to voices, and translate proprietary .x2s files to MIDI files.
 */
public class Transcriber {
    /* --- For translating / transposing --- */
    /*
      H-I talk about C, C', and C'', cf. p. 90, 94, but that does not make
      much sense compared to the remarks on transposition, cf. p. 154.
    */
    // This is H-I, p. 90:
    // private static int [] midiNotes = {48, 50, 52, 53, 55, 57, 59, 60, 62, 64, 65, 67, 69, 71, 72};
    // This is in accordance with the description of the transpositions,
    // p. 154:
    private static int [] midiNotes = {60, 62, 64, 65, 67, 69, 71, 72, 74, 76, 77, 79, 81, 83, 84};

    /* --- For MIDI file export --- */
    private static final int TIME_STEP = 24*1;  // resolution = 24 PPQ, therefore 1 quarter note = 24 pulses
    
    private Sequence seq;
    private int currentTime = 0;


    /* === Translating and Transposing Functions === */
    public static  int toMidiNote (int hiNote) {return midiNotes [hiNote];}


    public static int transposeNoteForVoice(int midiNote, int j, int voices)
    {
	// Transpose, depending on voice number.
	/*
	  Original H-I would be: violins by +12, viola by 0, cello by -24.
	  However, I don't know where I got this; cf. p. 154:
	  cello by -24, viola -12, violins +/-0.
	  This here is in accordance with the "midiNotes" defined above.
	*/
	if (j < voices / 4.0f) {  // cello
	    midiNote -= 24;
	}
	else if (j >= voices / 4.0f && j < voices / 2.0f) {  // viola
	    midiNote -= 12;
	}
	else if (j >= voices / 2.0f) {  // violin
	    midiNote += 0;
	}

	return midiNote;
    }


    public static int dynamicsToMidiVelocity(int d) {
	return 31 + 32*d;
    }


    /* === MIDI File Export Functions === */
    public Transcriber(int voices) throws InvalidMidiDataException {
	seq = new Sequence(Sequence.PPQ, 24, voices);

	// Set time
	MetaMessage mmsg = new MetaMessage();
	// 1,000,000 ms per quarter note:
	try{mmsg.setMessage(0x51, new byte[] {0x0f, 0x42, 0x40}, 3);}
	catch (Exception ex) {Log.write (Log.WARNING, ex.toString ());}
	MidiEvent mevt = new MidiEvent(mmsg, 0);
	seq.getTracks()[0].add(mevt);

	// Set programs to strings
	for(int j=0; j<voices; ++j) {
	    int program = 0;
	    ShortMessage msg = new ShortMessage();
	    
	    if (j < voices / 4.0f) program = 0x2a;  // cello
	    else if (j >= voices / 4.0f && j < voices / 2.0f) program = 0x29;  // viola
	    else if (j >= voices / 2.0f) program = 0x28;  // violin

	    Log.write(Log.DEBUG, "Setting program to " + program + " for voice " + j + ".");
	    try {msg.setMessage(ShortMessage.PROGRAM_CHANGE | (j & 0x0f), program, 0);}
	    catch (Exception ex) {Log.write (Log.WARNING, ex.toString ());}

	    MidiEvent evt = new MidiEvent(msg, 0);
	    seq.getTracks()[j].add(evt);
	}
    }


    public Sequence getSequence() {return seq;}

    
    public void addScore(Score sc) throws InvalidMidiDataException {
	int len = sc.getLength();
	int voices = sc.getNumVoices();  // This is not checked against the constructor.

	// For ligatures
	int[] lastNotes = new int[voices];
	int[] notesPlaying = new int[voices];

	for(int i=0; i<voices; ++i)
	    lastNotes[i] = -1;

	for(int i=0; i<len; ++i) {
	    for(int j=0; j<voices; ++j) {
		int hiNote = sc.getNote(j, i);  // (voice, chord)
		int midiNote = toMidiNote(hiNote);

		// Note on event
		midiNote = transposeNoteForVoice(midiNote, j, voices);

		int status, data1, data2;
		
		// Do a note on iff last note was different.
		if(midiNote != lastNotes[j]) {
		    status = ShortMessage.NOTE_ON | (j & 0x0f);
		    data1 = midiNote;
		    data2 = dynamicsToMidiVelocity(sc.getDynamics(j));
		    ShortMessage msgOn = new ShortMessage();
		    try {msgOn.setMessage (status, data1, data2);}
		    catch (Exception ex) {Log.write (Log.WARNING, ex.toString ());}

		    seq.getTracks()[j].add(new MidiEvent(msgOn, currentTime));

		    notesPlaying[j] = midiNote;
		    
		    // Log.write(Log.DEBUG, "Note ON : " + midiNote + " for voice " + j + ".");
		}
		
		// Note off event
		// Do a note off iff next note is different, or end of score.
		if(i<len-1 && notesPlaying[j] != Transcriber.transposeNoteForVoice(Transcriber.toMidiNote(sc.getNote(j, i+1)), j, voices) || i == len-1) {
		    status = ShortMessage.NOTE_OFF | (j & 0x0f);
		    data1 = midiNote;
		    data2 = 0;
		    ShortMessage msgOff = new ShortMessage();
		    try {msgOff.setMessage (status, data1, data2);}
		    catch (Exception ex) {Log.write (Log.WARNING, ex.toString ());}

		    int offTime = currentTime + TIME_STEP;
		    if(i == len-1) offTime += (int)(TIME_STEP / 4.0f);
		    seq.getTracks()[j].add(new MidiEvent(msgOff, offTime));
		    
		    // Log.write(Log.DEBUG, "Note OFF: " + midiNote + " for voice " + j + ".");
		}
	    }  // END for j (voices)
	    
	    // Remember last note
	    for(int j=0; j<voices; ++j)
		lastNotes[j] = notesPlaying[j];

	    // One step in time: advance time for next event
	    currentTime += TIME_STEP;

	    // Pause after one score.
	    if(i == len-1) currentTime += TIME_STEP * 2;
	}  // END for i (chords)
    }


    public void addScore(Score sc, String scoreInfo) throws InvalidMidiDataException {
	if(scoreInfo != null) {
	    MetaMessage mmsg = new MetaMessage();
	    Charset charset = StandardCharsets.US_ASCII;

	    byte[] data = charset.encode(scoreInfo).array();
	    mmsg.setMessage(0x01, data, data.length);
	    seq.getTracks()[0].add(new MidiEvent(mmsg, currentTime));
	}
	
	addScore(sc);
    }


    public void writeMidiFile(File outfile) throws IOException {
	MidiSystem.write(getSequence(), 1, outfile);
    }

    
    public static void main(String[] args) throws Exception {
	Log.setLevel(Log.DEBUG);

	System.out.println("Transcriber starts work on score(s)...");
	
	Vector<String> infiles = new Vector<String>();
	String outfile = "out.mid";

	// Process command line args
	for(int i=0; i<args.length; ++i) {
	    try {
		if(args[i].toUpperCase().equals("-O") || args[i].toUpperCase().equals("/O"))
		    outfile = args[++i];
		else infiles.add(args[i]);
	    }
	    catch(Exception ex) {
		Log.write(Log.ERROR, "Error retrieving file names.");
	    }
	}

	// Init classes
	Score sc = Librarian.fetch(new File(infiles.elementAt(0)));
	Transcriber tr = new Transcriber(sc.getNumVoices());

	// Add scores to MIDI sequence
	for(int i=0; i<infiles.size(); ++i) {
	    Log.write(Log.INFORMATION, "Adding score '" + infiles.elementAt(i) + "'...");

	    String filename = infiles.elementAt(i);
	    sc = Librarian.fetch(new File(filename));
	    String ruleSetStr = sc.getRuleSetStr();
	    tr.addScore(sc, ruleSetStr != null ? "Rule Set " + ruleSetStr : null);
	}

	// Write MIDI file
	Log.write(Log.INFORMATION, "Writing to output file '" + outfile + "'...");
	tr.writeMidiFile(new File(outfile));
    }
}
